@@ -1,7 +1,4 @@
 """ Tidy3d package imports"""
-from concurrent.futures import ProcessPoolExecutor, process
-
-from rich import pretty, traceback
 
 # grid
 from .components.grid.grid import Grid, Coords
@@ -13,6 +10,7 @@ from .components.geometry import Box, Sphere, Cylinder, PolySlab, GeometryGroup
 # medium
 from .components.medium import Medium, PoleResidue, AnisotropicMedium, PEC, PECMedium
 from .components.medium import Sellmeier, Debye, Drude, Lorentz
+from .components.medium import CustomMedium
 
 # structures
 from .components.structure import Structure, MeshOverrideStructure
@@ -92,11 +90,19 @@ from .log import log, set_logging_file, WebError
 from .config import config
 
 # version
-from .version import __version__
+from .version import __version__, PIP_NAME
 
 # updater
 from .updater import Updater
 
+# warn if this was uploaded to the beta PyPI, also sets the pyPI repository that we upload to
+if "beta" in PIP_NAME:
+    log.warning(
+        "This version of Tidy3D was pip installed from the 'tidy3d-beta' repository on PyPI. "
+        "Future releases will be uploaded to the 'tidy3d' repository. "
+        "From now on, please use 'pip install tidy3d' instead."
+    )
+
 
 def set_logging_level(level: str) -> None:
     """Raise a warning here instead of setting the logging level."""
@@ -106,7 +112,4 @@ def set_logging_level(level: str) -> None:
     )
 
 
-# make all stdout and errors pretty
-pretty.install()
-# traceback.install()
 log.info(f"Using client version: {__version__}")
